﻿namespace SpectralEstimation

open Exception
open UtilityFunctions
open System
open MathNet.Numerics

//Orthonormal linear spline basis on L^2[0,1]
type LinearSplinesBasis (dimension :int) =
    let J = dimension - 1
    let f i =
        if i = -1 then fun x -> 1.
        else 
            fun x -> if x >= (float i)/(float J) && x < (float i + 1.)/(float J) 
                     then (x- (float i)/(float J))
                     elif x < (float i)/(float J) then 0. 
                     else 1./ float J 
    let d i =
        if i = -1 then fun x -> 0.
        else           fun x -> if x >= (float i)/(float J) && x < (float i + 1.)/(float J) 
                                then float 1
                                else 0. 
    let functions   = Array.init dimension (fun i -> f <| i - 1)
    let derivatives = Array.init dimension (fun i -> d <| i - 1)
    interface IBasis with
        member this.IsOrthonormal = false
        member this.Dimension = dimension
        member this.GetFunction i = functions.[i]
        member this.Functions = functions
        member this.GetFunctionFromCoefficients coefficients =
            fun x -> Array.map2 (fun c f -> c * f x) coefficients functions   |> Array.sum
        member this.GetDerivativeFromCoefficients coefficients =
            fun x -> Array.map2 (fun c f -> c * f x) coefficients derivatives |> Array.sum

//Fourier cosine basis on L^2[0,1]
type FourierCosineBasis (dimension :int) =  
    let getF = function
        | 0 -> fun x -> 1.
        | i -> fun x -> (sqrt 2.) * cos (Math.PI * (float i) * x)
    let getD = function
        | 0 -> fun x -> 0.
        | i -> fun x -> - (sqrt 2.) * Math.PI * (float i) * sin (Math.PI * (float i) * x)
    let functions   = Array.init dimension getF
    let derivatives = Array.init dimension getD 
    interface IBasis with
        member this.IsOrthonormal = true
        member this.Dimension = dimension
        member this.GetFunction i = functions.[i]
        member this.Functions = functions
        member this.GetFunctionFromCoefficients coefficients =
            fun x -> Array.map2 (fun c f -> c * f x) coefficients functions   |> Array.sum
        member this.GetDerivativeFromCoefficients coefficients =
            fun x -> Array.map2 (fun c f -> c * f x) coefficients derivatives |> Array.sum