﻿namespace SpectralEstimation

open System
open MathNet.Numerics
open UtilityFunctions
open Exception
open Estimator

//Implementation of the Gobet, Hoffmann, Reiß volatility estimator,
//see Gobet, E., Homann, M., and Reiÿ, M. (2004)
//Nonparametric estimation of scalar diffusions based on low frequency data.
//The Annals of Statistics, 32:2223-2253.
type GHRVolatilityEstimator (sample :Sample, basis :IBasis) =
    //GHR estimator requires that the observations are equidistant
    let delta = match sample.Times with
                | Regular delta -> delta
                | Explicit ts -> failWith "GHR estimator requires a regular schedule of observation times"
    //Estimator of the occupation measure (stationary measure for T large)
    let mu = estimateMu sample basis
    //Estimator of the matrix g^hat
    let mtrxG = buildG sample basis
    //Estimator of the matrix l^hat = delta^{-1} (g^hat - p^hat)
    let mtrxL = buildL sample basis
    //Returns the first nontrivial eigenpair of the eigenproblem
    //  lambda g^hat x = delta^{-1} (g^hat - p^hat) x
    let lambda, x = solveGEP mtrxG mtrxL 1
    //To calculate the estimated eigenvalue of the infinitesimal generator recall that
    // p^hat x = kappa g^hat x.
    // kappa = 1. - delta * lambda = exp (delta * v)
    let v = log(1. - delta * lambda) / delta
    //We transform the eigenvector x into the estimated eigenfunction of the generator
    let u = basis.GetFunctionFromCoefficients x
    //and its derivative
    let uPrime = basis.GetDerivativeFromCoefficients x
    let volatility x =
        let int x = Integration.Integrate.OnClosedInterval((fun x -> u x * mu x), 0., x)
        2. * v * (int x) / (mu x * uPrime x)
    //Estimated volatility function
    member this.Volatility = volatility
    //Lp estimation error on the interval [a,b]
    member this.Error (p: float) (a :float) (b :float) (vol :float -> float) =
        calcLpNorm p a b (fun x -> volatility x - vol x)

//Implementation of the spectral estimator adapted to a random schedule of observation times,
//see Chorowski, J. and Trabs, M. (2016). Spectral estimation for diffusions with random sampling times
//Stochastic Processes and their Applications,  doi:10.1016/j.spa.2016.03.009
type ChTVolatilityEstimator (sample :Sample, basis :IBasis) =
    //Estimator of the occupation measure (stationary measure for T large)
    let mu = estimateMu sample basis
    //Estimator of the matrix g^hat
    let mtrxG = buildG sample basis
    //Estimator of the matrix (g^hat - p^hat)
    let mtrxGmP = buildGmP sample basis
    //Returns the first nontrivial eigenpair of the eigenproblem
    //  lambda g^hat x = (g^hat - p^hat) x
    let lambda, x = solveGEP mtrxG mtrxGmP 1
    //To calculate the estimated eigenvalue of the infinitesimal generator v
    //recall that
    // p^hat x = (1- lambda) g^hat x.
    // 1 - lambda = Laplace_gamma (v),
    // where Laplace_gamma is the Laplace transform of the time sampling distribution
    let v =
        //List of the time increments
        let tIncs = match sample.Times with
                    | Explicit ts -> Array.fold (fun (res, prev) next -> next - prev :: res, next)
                                        (List.empty, ts.[0]) (Array.sub ts 1 (ts.Length - 1)) |> fst
                    | Regular delta -> List.init sample.N (fun i -> delta)
        //Estimator of the Laplace_gamma
        let Laplace x = List.sumBy (fun t -> exp (- x * t)) tIncs / float tIncs.Length
        if lambda > 1. || lambda < 0. then failWith "Estimated eigenvalue is outside the range of the Laplace transform."
        InvertDecreasingFunction Laplace (System.Math.Pow(0.5,16.)) (1. - lambda)
    //We transform the eigenvector x into the estimated eigenfunction of the generator
    let u = basis.GetFunctionFromCoefficients x
    //and its derivative
    let uPrime = basis.GetDerivativeFromCoefficients x
    let volatility x =
        let int x = Integration.Integrate.OnClosedInterval((fun x -> u x * mu x), 0., x)
        -2. * v * (int x) / (mu x * uPrime x)
    //Estimated volatility function
    member this.Volatility = volatility
    //Lp estimation error on the interval [a,b]
    member this.Error (p: float) (a :float) (b :float) (vol :float -> float) =
        calcLpNorm p a b (fun x -> volatility x - vol x)   

//Implementation of the Florens-Zmirou estimator
//see Florens-Zmirou, D. (1993). On estimating the diffusion coeficient from discrete observations.
//Journal of Applied Probability, 30:790804
type FZVolatilityEstimator (sample :Sample, bandwidth :float) =
    //Array of squared increments [(X_{t_i+1}-X_{t_i})^2]
    let vals = Array.map2 (fun p n -> (n - p) * (n - p)) (Array.sub sample.Values 0 sample.N) (Array.sub sample.Values 1 sample.N)
    //An increment is used only if X_{t_i} is closer than the bandwidth value of x 
    let chars x = 
        let char v = if abs (v - x) < bandwidth
                        then 1.
                        else 0.
        Array.map char (Array.sub sample.Values 0 sample.N)

    let volatility x =
        //Identifies the increments relevant for the value of volatility at x
        let chars = chars x
        //Sums up the relevant increments, rescaled by the time t_i+1 - t_i
        let numerator = match sample.Times with
                        | Regular delta  -> (Array.sumBy (fun (c,v) -> c * v) <| Array.zip chars vals) / delta
                        | Explicit ts    -> //times between observations
                                            let incTs = Array.map2 (fun p n -> n - p) (Array.sub ts 0 sample.N) (Array.sub ts 1 sample.N)
                                            Array.sumBy (fun (t,c,v) -> c*v/t) (Array.zip3 ts chars vals)
        numerator / (Array.sum chars)
    member this.Volatility = volatility
    //Lp estimation error on the interval [a,b]
    member this.Error (p: float) (a :float) (b :float) (vol :float -> float) =
        calcLpNorm p a b (fun x -> volatility x - vol x)


//Implementation of the frequency universal estimator
//see Chorowski. Nonparametric volatility estimation in scalar diusions:
//Optimality across observation frequencies, arXive: http://arxiv.org/abs/1507.07139
type ChVolatilityEstimator (sample :Sample, basisDimension) =
    //Linear spline basis bith equidistant knots
    let basis = new LinearSplinesBasis (basisDimension) :> IBasis
    //Estimator requires that the observations are equidistant
    let delta = match sample.Times with
                | Regular delta -> delta
                | Explicit ts -> failWith "GHR estimator requires a regular schedule of observation times"
    //Estimator of the matrix g^hat
    let mtrxG = buildG sample basis
    //Estimator of the matrix l^hat = delta^{-1} (g^hat - p^hat)
    let mtrxL = buildL sample basis
    //Returns the first nontrivial eigenpair of the eigenproblem
    //  gamma g^hat x = l^hat x
    let gamma, x = solveGEP mtrxG mtrxL 1
    //Estimated eigenvalue of the infinitesimal generator divided by gamma
    let xi = if delta * gamma > 1. 
                then 0.
                else Math.Log(1. - delta * gamma) / (gamma * delta)
    //We transform the eigenvector x into the estimated eigenfunction of the generator
    let u = basis.GetFunctionFromCoefficients x
    //Derivative of the ith basis function
    let getPsiPrime i = 
        Array.init basis.Dimension (fun j -> if i=j then 1. else 0.)
            |> basis.GetDerivativeFromCoefficients
    //Estimated volatility is constant on intervals [(j-1)/J,j/J], j= 1,...,J
    //Following function returns the array of volatility values
    let sigma =
        let denominator i =
            //The value of the derivative of the eigenfunction on the interval [(j-1)/J,j/J]
            let derivative = x.[i]
            //Occupation measure of the interval [(j-1)/J,j/J] (note that PsiPrime is the characteristic function of [(j-1)/J,j/J])
            let occMeasure = Array.sumBy (fun x -> (getPsiPrime i) x) sample.Values / float sample.Values.Length
            derivative * occMeasure 
        let numerator i = Array.sumBy (fun j -> x.[j] * (mtrxL.ToArray()).[j,i]) (Array.init (basis.Dimension - 1) (fun i -> i + 1))
        Array.init (basis.Dimension - 1) (fun i -> -2. * xi * numerator (i+1) / denominator (i+1))
    let volatility x = Array.sumBy (fun i -> if (getPsiPrime <| i + 1) x = 1. then sigma.[i] else 0.) (Array.init (basis.Dimension - 1) (fun i -> i))
    member this.Volatility = volatility
    //Lp estimation error on the interval [a,b]
    member this.Error (p: float) (a :float) (b :float) (vol :float -> float) =
        calcLpNorm p a b (fun x -> volatility x - vol x)