﻿namespace SpectralEstimation

open Exception
open UtilityFunctions
open System
open MathNet.Numerics


module SampleGenerator =

    let rec private reflect x = 
        if x < 0. then -x |> reflect
        elif x > 1. then 2. - x |> reflect
        else x

    let private counterArray N = Array.init N (fun i -> i)

    let generateRegularSample (data :DiffusionData) delta N :Sample =
        let dt = min (delta/100.) 0.001
        let k = Math.Ceiling (delta/dt)
        let ddt = delta / k
        let rnd = new Random.Mcg31m1() 
        let inc = 
            let gen = Distributions.Normal.Samples (rnd, 0. , sqrt ddt)
            gen.GetEnumerator()
        inc.MoveNext() |> ignore
        let counter = counterArray (N+101)
        let getNext (x :float) =
            let y = x +                                       //current state
                     ddt * data.Drift x +                     //drift
                     sqrt (data.Volatility x) * inc.Current   //volatility
                     |> reflect                               //reflect at boundaries
            inc.MoveNext() |> ignore
            y
        let getNextK x = Array.fold (fun p c -> getNext p) x (counterArray <| int k)
        let sample = Array.fold (fun sample i -> match sample with
                                                    | h :: t -> (getNextK h) :: sample
                                                    | _ -> [0.6]) //Starting value
                                                    List.empty counter
        inc.Dispose()
        let values = Array.sub  (sample |> List.rev |> Array.ofList) 99 (N+1)
        {   
            Values   = values
            Times    = Regular delta
        }

    //Generates equidistant observations with the occupation density uniformly above a prescribed level v
    let rec generateRegularSampleWithMinOccDensity diffusionData delta (T :float) v =
        //generate candidate sample
        let sample = generateRegularSample diffusionData delta (int <| T/delta)
        //estimate the occupation density
        let mu    = Estimator.estimateMu sample (new FourierCosineBasis 30 :> IBasis)
        //check if the occupation density os above the prescribed level,
        //if not, generate a new sample
        if minF mu >= v
            then sample 
            else generateRegularSampleWithMinOccDensity diffusionData delta T v

    let private DrawRegularSchedule delta N = Explicit <| Array.init (N+1) (fun i -> float i * delta)
    
    //Sampling times are distributed exponentially with intensity 1,/mean
    let private DrawExponentialSchedule mean N =
        let seed = System.DateTime.Now.Millisecond
        let generator = System.Random(seed)
        let gen =
            let dist = (MathNet.Numerics.Distributions.Exponential(1./mean)).Samples()
            dist.GetEnumerator()
        let ts = Array.init N (fun i -> gen.MoveNext() |> ignore
                                        gen.Current)
        gen.Dispose()
        Explicit <| Array.scan (fun sum t -> sum + t) 0. ts

    //Sampling times are distributed uniformly on the interval [0,2*delta]
    let private DrawUniformSchedule delta N =
        let seed = System.DateTime.Now.Millisecond
        let generator = System.Random(seed)
        let gen =
            let dist = (MathNet.Numerics.Distributions.ContinuousUniform()).Samples()
            dist.GetEnumerator()
        let ts = Array.init N (fun i -> gen.MoveNext() |> ignore
                                        2. * delta * gen.Current)
        gen.Dispose()
        Explicit <| Array.scan (fun sum t -> sum + t) 0. ts

    //Sampling times follow the Beta(alpha,alpha) distribution rescaled to the interval [0, 2*delta]
    let private DrawBetaSchedule delta alpha N =
        let seed = System.DateTime.Now.Millisecond
        let generator = System.Random(seed)
        let gen =
            let dist = (MathNet.Numerics.Distributions.Beta (alpha, alpha)).Samples()
            dist.GetEnumerator()
        let ts = Array.init N (fun i -> gen.MoveNext() |> ignore
                                        2. * delta * gen.Current)
        gen.Dispose()
        Explicit <| Array.scan (fun sum t -> sum + t) 0. ts

    //Wrapper function over different sampling distributions
    let generateRandomSchedule (distribution :RandomScheduleType) (length :int) =
        match distribution with
        | RandomScheduleType.Deterministic delta   -> DrawRegularSchedule       delta     length
        | RandomScheduleType.Exponential mean      -> DrawExponentialSchedule   mean      length
        | RandomScheduleType.Uniform delta         -> DrawUniformSchedule       delta     length
        | RandomScheduleType.Beta (delta,alpha)    -> DrawBetaSchedule          delta     alpha     length

    //Draws a sample with predefined observation times.
    let generateSampleFromTimeSchedule (data :DiffusionData) (schedule :TimeSchedule) =
        let rnd = new Random.Mcg31m1() 
        let inc = 
            let gen = Distributions.Normal.Samples (rnd, 0. , 1.)
            gen.GetEnumerator()
        inc.MoveNext() |> ignore
        //We use Euler scheme to generate the value of the process at time t_1
        let getNext (x_0 :float) (t_0 :float) (t_1 :float) =
            let dt = min (0.01 * abs (t_1 - t_0)) 0.001 
            let moveByDt x =
                let y = x +                                               //current state
                            dt * data.Drift x +                           //drift
                            sqrt (dt * data.Volatility x) * inc.Current   //volatility
                inc.MoveNext() |> ignore
                y |> reflect                                              //reflect at boundaries
            let mutable t = t_0
            let mutable x = x_0
            while t < t_1 do
                x <- moveByDt x
                t <- t + dt
                ()
            x
        //Observation times
        let times = match schedule with
                    | Regular delta -> failWith "Observation times must be specified"
                    | Explicit times -> times
        let values = Array.fold (fun (t_0 ,path) t_1 -> match path with
                                                        | h :: t -> t_1, getNext h t_0 t_1 :: path
                                                        | _ -> failWith "The initial value should be already specified")
                            (times.[0], [0.6]) (Array.sub times 1 (times.Length-1)) 
        inc.Dispose()
        {   Values = values |> snd |> List.rev |> List.toArray
            Times  = schedule }