﻿namespace SpectralEstimation

exception Error of string

module Exception = 
    let failWith msg = raise (Error msg)

//Provides access to diferent L^2 basis on R
//A basis is a finite colection of linearly independent functions,
//not necessarily orthonormal.
type IBasis =
    abstract IsOrthonormal                          :bool
    //Dimension of the spanned subspace
    abstract Dimension                              :int
    //Returns basis function of given index
    abstract GetFunction                            :int -> (float -> float)
    //Returns an array consisitng of all functions
    abstract Functions                              :(float -> float) []
    //Converts an array of coefficients into the corresponding spanned function 
    abstract GetFunctionFromCoefficients            :float [] -> (float -> float)
    //Converts an array of coefficients into the derivative of the corresponding spanned function
    //Method makes sense only if basis functions are differentiable
    abstract GetDerivativeFromCoefficients          :float [] -> (float -> float)


//Contains the information about observation times.
type TimeSchedule =
    | Regular of float      //equidistant times of observations
    | Explicit of float []  //array of observation times
    
//Types of random times schedules
type RandomScheduleType =
    | Deterministic of float
    | Exponential   of float
    | Beta          of float * float
    | Uniform       of float

//Contains the information about diffusion coefficients
type DiffusionData =
    {
        Volatility :float -> float
        Drift      :float -> float
    }


//Contains information about the observations of the diffusion process
//In case of the explicit schedule the length of the time schedule
// should match the length of the observations array!
type Sample = 
    {
        Times   :TimeSchedule
        Values  :float []
    }
    //Returns the time horizon (time of the last observation)
    //Note that in case of Regular schedule the first observation is at time 0 
    member this.T = match this.Times with 
                    | Regular delta -> delta * float (this.Values.Length - 1)
                    | Explicit times -> times.[times.Length-1]
    //Index of the last observation
    member this.N = this.Values.Length - 1