﻿namespace SpectralEstimation

open MathNet.Numerics
open System
open Exception

//Functions of general utility
module UtilityFunctions = 

    //Returns the approximated minimal value of a function on [0,1]
    let minF (f :float -> float) = Array.init 999 (fun i -> f (float (i+1)/1000.)) |> Array.min

    //Calculates the L^p norm on [a,b] of function f
    let calcLpNorm (p :float) (a :float) (b :float) (f :float -> float) =
        let int = Integration.Integrate.OnClosedInterval ((fun x -> Math.Pow (f x |> abs, p)), a, b)
        Math.Pow(int, 1./p)

    //Inverts a decreasing function
    let InvertDecreasingFunction (f : float -> float) (precision :float) y =
        let mutable l = 0.
        let mutable r = 1.
        while f r > y do l <- r
                         r <- r + 1.
        while r - l > precision do if f ((l + r) / 2.) > y then l <- (l + r) / 2.
                                                           else r <- (l + r) / 2.
        (l + r) / 2.

    //Verifies if a is a multiple of b
    let private isMult (a :int) (b :int) = if a % b = 0 then true else false

    //Divides a single array of length n * dim into n lists of length dim
    let private toVectors (ar :float []) (dim :int) =
        if not (isMult ar.Length dim)
            then failWith "Input array's length %i should be a multiple of dimension dim = %i." ar.Length dim
        Array.fold (fun (vec, i) el ->
                        if i = 0
                            then ([el] :: vec, i + 1)
                        else 
                             match vec with
                             | h :: tail -> (el :: h) :: tail, if i + 1 < dim then i + 1 else 0
                             | _ -> failWith "vec should not be empty as it is populated in the first step!")
                    (List.empty, 0) ar |> fst |> Array.ofList |> Array.map List.rev |> Array.rev  |> Array.map List.toArray

    //Returns the i th eigenpair of the generalized eigenproblem
    //                 lambda Ax=Bx
    //where A and B are real symmetric square matrices of the same dimension.
    //Function does not check the correctness of the input data!
    let solveGEP (A :LinearAlgebra.Generic.Matrix<float>) (B :LinearAlgebra.Double.DenseMatrix) (i :int) =
        let dim = A.ColumnCount
        //Calculates the matrix A^{-1}B
        let BAInv = 
            let AInv = A.Inverse()
            AInv.Multiply B
        //We use MathNet library to solve the eigenproblem
        //             lambda x = A^{-1}Bx
        let eigenValues, eigenVectors = 
            //Empty containers for results
            let eigenValues :System.Numerics.Complex [] = Array.init (dim) (fun i -> System.Numerics.Complex.Zero)
            let eigenVectors = Array.init (dim*dim) (fun i -> 0.)
            let matrixD = Array.init (dim*dim) (fun i -> 0.)
            //Execution of the algorithm
            let algebra = new MathNet.Numerics.Algorithms.LinearAlgebra.ManagedLinearAlgebraProvider() :> Algorithms.LinearAlgebra.ILinearAlgebraProvider
            algebra.EigenDecomp(BAInv.IsSymmetric, dim, BAInv.ToColumnWiseArray(), eigenVectors, eigenValues, matrixD)
            //Eigenvectors are concatenated in the form of a dim*dim array
            Array.zip 
                (Array.map (fun (z :System.Numerics.Complex) -> z.Real ) eigenValues)
                (toVectors eigenVectors dim)
                |> Array.sortBy (fun x -> fst x)
                |> Array.unzip
        eigenValues.[i], eigenVectors.[i]

//Building blocks for the estimators
module Estimator =

    //Constructs mu^hat - estimator of the occupation time measure
    let estimateMu (sample :Sample) (basis :IBasis) =
        //Estimates the value of the scalar product <psi,mu>
        let calc (psi :float -> float) =
            //Underweighting of the first and last observation
            //so that we are in line with the definition of the gram matrix
            let weight i = if i = 0 || i = sample.N then 0.5 else 1.
            //<psi,mu>~sum psi(X_i)/N
            let sum = Array.mapi (fun i x -> (weight i ) * psi x) sample.Values |> Array.sum
            sum/(float sample.N)        
        if basis.IsOrthonormal
            //When basis is orthonormal the scalar product <psi,mu>
            //is the coefficient in the basis expansion.
            then Array.map calc basis.Functions |> basis.GetFunctionFromCoefficients
            //If not, we have to calculate the coefficients using the Gram matrix of the basis.
            else let fs = MathNet.Numerics.LinearAlgebra.Double.DenseVector.Create(basis.Dimension, fun i -> basis.Functions.[i] |> calc)
                 let gram = MathNet.Numerics.LinearAlgebra.Double.DenseMatrix.Create
                             (basis.Dimension, basis.Dimension, fun i j -> MathNet.Numerics.Integration.Integrate.OnClosedInterval ((fun x -> basis.Functions.[i] x * basis.Functions.[j] x), 0., 1.))
                 let cfs = (gram.Inverse()).LeftMultiply fs
                 cfs.ToArray() |> basis.GetFunctionFromCoefficients

    //Constructs the Gramm matrix g^hat
    let buildG (sample :Sample) (basis :IBasis) =
        let mtrx =
            let weight n = if n = 0 || n = sample.N
                                then sqrt (0.5/( float sample.N))
                                else 1. / sqrt(float sample.N)
            LinearAlgebra.Double.DenseMatrix.Create(basis.Dimension, sample.N + 1, 
                fun j n -> (weight n) * (basis.GetFunction j) sample.Values.[n])
        mtrx.TransposeAndMultiply mtrx

    //Constructs the matrix  l^hat = (g^hat - p^hat) / delta
    let buildL (sample :Sample) (basis :IBasis) =
        //defines the [j,n] entry of the matrix mtrx
        let init =
            new Func<int,int,float>
                (fun j n ->
                    let psi = basis.GetFunction j
                    psi sample.Values.[n+1] - psi sample.Values.[n])
        let mtrx = LinearAlgebra.Double.DenseMatrix.Create(basis.Dimension, sample.N, init)
        (mtrx.TransposeAndMultiply mtrx).Divide (2. * float sample.T) :?> MathNet.Numerics.LinearAlgebra.Double.DenseMatrix

    //Constructs the matrix  (g^hat - p^hat)
    let buildGmP (sample :Sample) (basis :IBasis) =
        //defines the [j,n] entry of the matrix mtrx
        let init =
            new Func<int,int,float>
                (fun j n ->
                    let psi = basis.GetFunction j
                    psi sample.Values.[n+1] - psi sample.Values.[n])
        let mtrx = LinearAlgebra.Double.DenseMatrix.Create(basis.Dimension, sample.N, init)
        (mtrx.TransposeAndMultiply mtrx).Divide (2. * float sample.N) :?> MathNet.Numerics.LinearAlgebra.Double.DenseMatrix

open System.Windows.Forms
open System.Windows.Forms.DataVisualization.Charting

type DataChartForm(title, data : (float * float) [] ) =
    inherit Form( Text=title )
    let chart = new Chart(Dock=DockStyle.Fill)
    let area = new ChartArea(Name="Area1")
    let series = new Series()
    do series.ChartType <- SeriesChartType.Line
    do data |> Seq.iter (series.Points.AddXY >> ignore)
    do series.Color <- System.Drawing.ColorTranslator.FromHtml "black"
    do series.BorderWidth <- 3
    do chart.Series.Add( series )
    do chart.ChartAreas.Add(area)
    do base.Controls.Add( chart )

type ComparisonChartForm( title, xs : (float * float) [], f : float -> float) =
    inherit Form( Text=title )

    let chart = new Chart(Dock=DockStyle.Fill)
    let area = new ChartArea(Name="Area1")
    
    let series = new Series()
    do series.ChartType <- SeriesChartType.FastLine
    do xs |> Seq.iter (series.Points.AddXY >> ignore)
    do series.ChartArea <- "Area1"
    do series.BorderWidth <- 3
    do series.Color <- System.Drawing.ColorTranslator.FromHtml "red"
    do chart.Series.Add( series )

    let vs = Array.map (fun (x,y) -> (x, f x)) xs
    let series = new Series()
    do series.ChartType <- SeriesChartType.FastLine
    do vs |> Seq.iter (series.Points.AddXY >> ignore)
    do series.ChartArea <- "Area1"
    do series.BorderWidth <- 3
    do series.Color <- System.Drawing.ColorTranslator.FromHtml "blue"
    do chart.Series.Add( series )

    do chart.ChartAreas.Add(area)
    do base.Controls.Add( chart )

//Utility functions for data display
module Plot =

    let Sample (sample :Sample) =
        let ts = match sample.Times with
                 | Explicit ts -> ts
                 | Regular delta -> Array.init sample.Values.Length (fun i -> delta * float i)
        let chart = new DataChartForm ("Sample", Array.zip ts sample.Values)
        chart.Show()

    let private xs ( f:float -> float) (x_start :float) (x_end :float) (dx :float) = 
        Array.init (int ((x_end - x_start)/ dx)) (fun i -> let x = x_start + (float i) * dx
                                                           x, f x)

    let Function (f: float -> float) =
        let chart = new DataChartForm ("", xs f 0. 1. 0.001)
        chart.Show()

    let Compare (f_1 :float -> float) (f_2 :float -> float) =
        let chart = new ComparisonChartForm("", xs f_1 0. 1. 0.001, f_2)
        chart.Show()